/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */

import Foundation

/**
 This class represents the events to which can be subscribed and published.
 It follows the factory pattern and should be a singleton.
 This is implemented by having a sharedInstance variable which contains a dispatch_once_t token.
 This creates a static instance of Event exactly once, and returns this all following times. This means the event initialization is lazy.
 */

struct HandlerMap {
    var id: String
    var handler: EventHandler
}

@objc public class BlueConicEventManager: NSObject {

    // We provide a dictionary containing event names and their corresponding  handlerobjects and handlerfunctions, these are wrapped in an array since
    // one event name might have to respond to several handler objects and functions(rules in blueconic)
    internal var handlersByEvent: [String: [HandlerMap]] = [:]

    // all published events are stored, if the subscribe event is done later these can still be published
    @objc internal var eventQueue: [Event] = []

    /**
     Publish an event to all subscribers, add any parameters as a String array. These will be used to match subscribers if configured in Blueconic.
     - parameter event: The event that will be published
     */
    @objc public func publish(_ event: Event) {
        event.location = BlueConicClient.getInstance(nil).getScreenName()
        self.eventQueue.append(event)
        self.handleEvent(event)
    }

    /**
     * Clears the event queue
     */
    @objc public func clearEvents() {
        self.eventQueue.removeAll()
    }

    /**
     Clears the event handlers, takes an interaction id that matches with several rule id's, removes those rule id's
     - parameter key: To clear the event handlers is based on the key
     */
    @objc public func clearEventHandlers(_ key: String?) {
        if let validKey = key {
            for (className, handlerMapList) in self.handlersByEvent {
                for (index, element) in handlerMapList.enumerated() where element.id == validKey {
                    self.handlersByEvent[className]?.remove(at: index)
                    if self.handlersByEvent[className]?.count == 0 {
                        self.handlersByEvent.removeValue(forKey: className)
                    }
                }
            }

        } else {
            self.handlersByEvent = [:]
        }
    }

    /**
     Clears the event when the location is changed (another activity is opened)
     */
    @objc public func cleanup() {
        let location = BlueConicClient.getInstance(nil).getScreenName()
        var removeList: [Int] = []
        for (index, event) in eventQueue.enumerated() where location != event.location {
            removeList.append(index)
        }
        for index in removeList.sorted(by: >) {
            self.eventQueue.remove(at: index)
        }
    }

    /**
     Handles an event after publishing.

     - parameter event: A published event.
     */
    @objc internal func handleEvent(_ event: Event?) {
        for queuedEvent in self.eventQueue {
            if event == nil || event! === queuedEvent {
                if let eventHandlers = self.handlersByEvent[queuedEvent.name] {
                    for handlerMap in eventHandlers {
                        let curHandler = handlerMap.handler.listenerId
                        if !queuedEvent.handledBy.contains(curHandler) {
                            queuedEvent.addHandledBy(curHandler)
                            handlerMap.handler.handleEvent(queuedEvent)
                        }
                    }
                }
            }
        }
    }

    /**
     Adds an event listener. This listens to publish events with the same name and optionally a certain eventObject configuration.

     - parameter eventName:  A string containg the publish event name we should listen to
     - parameter callbackObject: A function describing how a handled event should be...handled.
     - parameter uuid:  A UUID for the listener that subscribed
     */
    internal func subscribe(_ eventName: String, callbackObject: EventHandler, uuid: String) {
        let name = eventName

        if var handlers = self.handlersByEvent[name] {
            // loop over handlers to make sure they're unique
            var addHandler = true
            // if a handler already occurs, break and don't save this handler into the handlersByEvent list
            for handlerMap in handlers where handlerMap.handler.listenerId == uuid {
                addHandler = false
                break
            }
            if addHandler {
                handlers.append(HandlerMap(id:uuid, handler: callbackObject))
                self.handlersByEvent[name] = handlers
            }
        } else {
            self.handlersByEvent[name] = [HandlerMap(id:uuid, handler: callbackObject)]
        }
        handleEvent(nil)
    }
}
